/**
 *
 * Compiz galaxy livewallpaper plugin
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright (C) 2011 Maximilian Schnarr <Fyrmir@gmx.de>
 *
 *
 * Several parts taken from the original plugin developed by Pal Dorogi
 * with the following copyright:
 *
 * Copyright (C) 2010 Pal Dorogi <pal.dorogi@gmail.com>
 *
 * Original plugin can be found here:
 * http://launchpad.net/compiz-plugins-livewallpapers
 *
 **/

#include <cmath>

#include "galaxy.h"

#define _PI (float) 3.141593

#define FOVY 45.0f
#define NEAR  0.1f
#define FAR  20.0f

COMPIZ_PLUGIN_20090315(galaxy, GalaxyPluginVTable);

/*****************************
 * Fake desktop window methods
 *****************************/
/*
 * Completely taken from compiz's wallpaper plugin.
 * http://cgit.compiz.org/compiz/plugins/wallpaper/
 */
static Visual *
findArgbVisual(Display *dpy,
		int screen)
{
	XVisualInfo *xvi;
	XVisualInfo temp;
	int nvi;
	int i;
	XRenderPictFormat *format;
	Visual *visual;

	temp.screen = screen;
	temp.depth = 32;
	temp.c_class = TrueColor;

	xvi = XGetVisualInfo(dpy, VisualScreenMask | VisualDepthMask |
			VisualClassMask, &temp, &nvi);
	if(!xvi)
		return 0;

	visual = 0;
	for(i = 0; i < nvi; i++)
	{
		format = XRenderFindVisualFormat(dpy, xvi[i].visual);
		if(format->type == PictTypeDirect && format->direct.alphaMask)
		{
			visual = xvi[i].visual;
			break;
		}
	}

	XFree(xvi);

	return visual;
}

void
GalaxyScreen::createFakeDesktopWindow()
{
	Display *dpy = screen->dpy();
	XSizeHints xsh;
	XWMHints xwmh;
	XSetWindowAttributes attr;
	Visual *visual;
	XserverRegion region;

	visual = findArgbVisual(dpy, screen->screenNum ());
	if(!visual)
		return;

	xsh.flags = PSize | PPosition | PWinGravity;
	xsh.width = 1;
	xsh.height = 1;
	xsh.win_gravity = StaticGravity;

	xwmh.flags = InputHint;
	xwmh.input = 0;

	attr.background_pixel = 0;
	attr.border_pixel = 0;
	attr.colormap = XCreateColormap(dpy, screen->root (),
			visual, AllocNone);

	fakeDesktop = XCreateWindow(dpy, screen->root (), -1, -1, 1, 1, 0, 32,
				InputOutput, visual,
				CWBackPixel | CWBorderPixel | CWColormap,
				&attr);

	XSetWMProperties(dpy, fakeDesktop, NULL, NULL,
				programArgv, programArgc, &xsh, &xwmh, NULL);

	XChangeProperty(dpy, fakeDesktop, Atoms::winStateSkipPager,
				XA_ATOM, 32, PropModeReplace,
				(unsigned char *) &Atoms::winStateSkipPager, 1);

	XChangeProperty(dpy, fakeDesktop, Atoms::winType,
				XA_ATOM, 32, PropModeReplace,
				(unsigned char *) &Atoms::winTypeDesktop, 1);

	region = XFixesCreateRegion(dpy, NULL, 0);

	XFixesSetWindowShapeRegion(dpy, fakeDesktop, ShapeInput, 0, 0, region);

	XFixesDestroyRegion(dpy, region);

	XMapWindow(dpy, fakeDesktop);
	XLowerWindow(dpy, fakeDesktop);
}

void
GalaxyScreen::destroyFakeDesktopWindow()
{
	if(fakeDesktop != None)
		XDestroyWindow(screen->dpy (), fakeDesktop);

	fakeDesktop = None;
}

/*****************
 * Drawing methods
 *****************/
void
GalaxyScreen::drawSpace()
{
	if(!spaceTexture.size())
		return;

	/* Load space matrix. */
	glPushMatrix();
	glLoadMatrixf( spaceMatrix.getMatrix() );

	/* Enable texture and draw plane. */
	spaceTexture[0]->enable(GLTexture::Good);

	glBegin(GL_QUADS);
		glTexCoord2d(1.0f, 0.0f);
		glVertex2f(-1.0f, -1.0f);

		glTexCoord2d(0.0f, 0.0f);
		glVertex2f(1.0f, -1.0f);

		glTexCoord2d(0.0f, 1.0f);
		glVertex2f(1.0f, 1.0f);

		glTexCoord2d(1.0f, 1.0f);
		glVertex2f(-1.0f, 1.0f);
	glEnd();

	/* Disable texture. */
	spaceTexture[0]->disable();

	/* Restore previous matrix. */
	glPopMatrix();
}

void
GalaxyScreen::drawLight()
{
	if(!lightTexture.size())
		return;

	lightTexture[0]->enable(GLTexture::Good);

	glBegin(GL_QUADS);
		glTexCoord2d(1.0f, 0.0f);
		glVertex2f(-1.0f, -1.0f);

		glTexCoord2d(0.0f, 0.0f);
		glVertex2f(1.0f, -1.0f);

		glTexCoord2d(0.0f, 1.0f);
		glVertex2f(1.0f, 1.0f);

		glTexCoord2d(1.0f, 1.0f);
		glVertex2f(-1.0f, 1.0f);
	glEnd();

	lightTexture[0]->disable();
}

void
GalaxyScreen::drawGalaxy()
{
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE); /* Additive */

	/* Reset and save the coordinate system before touching. */
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();

	/*
	 * Using glFrustum instead of gluPrespective.
	 * http://www.opengl.org/wiki/GluPerspective_code
	 */
	double ymax, xmax;
	ymax = NEAR * tanf (FOVY * _PI / 360.0);
	xmax = ymax * ((double) screen->width() / (double) screen->height());
	/* Set the correct Galaxy perspective. */
	glFrustum(-xmax, xmax, -ymax, ymax, NEAR, FAR);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();

		/* Draw background. */
		drawSpace();

		/* Apply the user defined transformations. */
		glLoadMatrixf( matrix.getMatrix() );

		/* Draw particles and light. */
		ps->drawParticles( optionGetStarByStar() );
		drawLight();

	/* Restore the coordinate system. */
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	/* Restore the original perspective. */
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	/* Restore blendig and other stuffs. */
	glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
	glDisable(GL_BLEND);
	glColor4usv(defaultColor);
}

void
GalaxyScreen::drawGalaxyToTexture()
{
	/* viewport has the same size as the texture */
	glViewport(0, 0, screen->width(), screen->height());

	/* activate buffer */
	if(GL::fbo)
		GL::bindFramebuffer(GL_FRAMEBUFFER_EXT, galaxyBuffer);
	else
	{
		/* save draw and read buffer */
		glPushAttrib(GL_COLOR_BUFFER_BIT | GL_PIXEL_MODE_BIT);

		glDrawBuffer(GL_BACK);
		glReadBuffer(GL_BACK);
	}

	/* draw galaxy */
	glClearColor(0, 0, 0, 0);
	glClear(GL_COLOR_BUFFER_BIT);
	drawGalaxy();

	/* finish buffer */
	if(GL::fbo)
		GL::bindFramebuffer(GL_FRAMEBUFFER_EXT, 0);
	else
	{
		glBindTexture(GL_TEXTURE_2D, galaxyTexture[0]->name());
		glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, 0, 0,
			screen->width(), screen->height());
		glBindTexture(GL_TEXTURE_2D, 0);

		/* restore the draw and read buffer */
		glPopAttrib();
	}
}

/**************
 * Screen hooks
 **************/
void
GalaxyScreen::preparePaint(int msSinceLastPaint)
{
	if(state == GalaxyScreen::GalaxyTrue)
		ps->updateParticles( optionGetSpeedRatio() * msSinceLastPaint );

	cScreen->preparePaint(msSinceLastPaint);
}

bool
GalaxyScreen::glPaintOutput (const GLScreenPaintAttrib &attrib,
		const GLMatrix &mtrx,
		const CompRegion &region,
		CompOutput *output,
		unsigned int mask)
{
	if(state == GalaxyScreen::GalaxyTrue)
		drawGalaxyToTexture();

	return gScreen->glPaintOutput(attrib, mtrx, region, output, mask);
}

void
GalaxyScreen::donePaint()
{
	if(state == GalaxyScreen::GalaxyTrue)
		cScreen->damageScreen();

	cScreen->donePaint();
}

void
GalaxyScreen::handleEvent(XEvent *event)
{
	screen->handleEvent(event);

	if(!screen->desktopWindowCount() && fakeDesktop == None)
	{
		createFakeDesktopWindow();

		if(state != GalaxyScreen::GalaxyUndefined)
			state = GalaxyScreen::GalaxyTrue;
	}

	if(screen->desktopWindowCount() > 1 && fakeDesktop != None)
	{
		destroyFakeDesktopWindow();

		if(state != GalaxyScreen::GalaxyUndefined)
			state = GalaxyScreen::GalaxyFalse;
	}
}

/**********************
 * Toggle key callbacks
 **********************/
bool
GalaxyScreen::initGalaxy(CompAction* action,
		CompAction::State actionState,
		CompOption::Vector& options)
{
	if(state == GalaxyScreen::GalaxyTrue)
		termGalaxy(action, actionState, options);
	else if(state != GalaxyScreen::GalaxyUndefined)
		state = GalaxyScreen::GalaxyTrue;

	return true;
}

bool
GalaxyScreen::termGalaxy(CompAction* action,
		CompAction::State actionState,
		CompOption::Vector& options)
{
	if(state != GalaxyScreen::GalaxyUndefined)
		state = GalaxyScreen::GalaxyFalse;

	return true;
}

/**************************
 * Private attribute helper
 **************************/
void
GalaxyScreen::loadLightTexture()
{
	CompString file = CompString("galaxy-light.png");
	CompString pname = CompString("galaxy");

	lightTexture = GLTexture::readImageToTexture(file, pname, lightSize);

	/*
	 * It is enough to show this message here. The drawing method
	 * will handle it correctly.
	 */
	if(!lightTexture.size())
		compLogMessage("galaxy", CompLogLevelWarn, "No light texture!");
}

void
GalaxyScreen::loadSpaceTexture()
{
	CompString file = CompString("space.png");
	CompString pname = CompString("galaxy");

	spaceTexture = GLTexture::readImageToTexture(file, pname, spaceSize);

	/*
	 * It is enough to show this message here. The drawing method
	 * will handle it correctly.
	 */
	if(!spaceTexture.size())
		compLogMessage("galaxy", CompLogLevelWarn, "No space texture!");

	/* create space matrix */
	spaceMatrix.reset();
	spaceMatrix.translate(0.0f, 0.0f,
		-1.5f * ((GLfloat) screen->width() / (GLfloat) screen->height()));
}

void
GalaxyScreen::updateMatrix()
{
	matrix.reset();

	matrix.translate(optionGetOffsetX(), optionGetOffsetY(),
		-optionGetZoom() * ((GLfloat) screen->width() / (GLfloat) screen->height()));

	matrix.rotate(optionGetRotateX(), 1.0f, 0.0f, 0.0f);
	matrix.rotate(optionGetRotateY(), 0.0f, 1.0f, 0.0f);
	matrix.rotate(optionGetRotateZ(), 0.0f, 0.0f, 1.0f);
}

/*************************
 * Option handling methods
 *************************/
void
GalaxyScreen::particlesOptionChanged(CompOption *o, Options num)
{
	switch(num)
	{
		case GalaxyOptions::StarsCount:
			try
			{
				ps->setStarsCount( optionGetStarsCount() );
			}
			catch (std::bad_alloc&)
			{
				/*
				 * At this point, it should be impossible to reactivate
				 * the galaxy except the reallocation succeeded.
				 */
				compLogMessage("galaxy", CompLogLevelFatal,
					"Cannot allocate enough memory for particle system!");

				state = GalaxyScreen::GalaxyUndefined;
				break;
			}

			/*
			 * Reenable the activation of the galaxy.
			 * The allocation succeeded.
			 */
			if(state == GalaxyScreen::GalaxyUndefined)
				state = GalaxyScreen::GalaxyFalse;

			if(fakeDesktop != None)
				state = GalaxyScreen::GalaxyTrue;

			break;

		case GalaxyOptions::StarsSize:
			ps->setStarsSize( optionGetStarsSize() );
			break;

		case GalaxyOptions::DrawStreaks:
			ps->generateFlare( optionGetDrawStreaks() );
			break;

		default:
			break;
	}
}

void
GalaxyScreen::matrixOptionChanged(CompOption *o, Options num)
{
	updateMatrix();

	if(state == GalaxyScreen::GalaxyTrue)
		cScreen->damageScreen();
}

/***********************
 * Window drawing method
 ***********************/
bool
GalaxyWindow::glDraw(const GLMatrix &transform,
		GLFragment::Attrib &attrib,
		const CompRegion &region,
		unsigned int mask)
{
	GALAXY_SCREEN(screen);

	if((gs->state == GalaxyScreen::GalaxyTrue || gs->fakeDesktop != None)
			&& window->type() & CompWindowTypeDesktopMask)
	{
		/*
		 * We let compiz draw a window with our texture covering
		 * the whole screen.
		 */
		GLTexture::MatrixList matrixList(1);
		matrixList[0] = gs->galaxyTexture[0]->matrix();

		/* flip the texture vertically */
		matrixList[0].y0 += screen->height() * matrixList[0].yy;
		matrixList[0].yy *= -1.0f;

		/* draw the window */
		gWindow->geometry().reset();
		gWindow->glAddGeometry(matrixList, screen->region(), infiniteRegion);

		if(gWindow->geometry().vCount)
			gWindow->glDrawTexture(gs->galaxyTexture[0], attrib, mask);

		/* We won't draw the real desktop window, we replace it. */
		return true;
	}

	return gWindow->glDraw(transform, attrib, region, mask);
}

/****************************
 * Constructor and Destructor
 ****************************/
GalaxyScreen::GalaxyScreen(CompScreen* s):
	PluginClassHandler<GalaxyScreen, CompScreen>(s),
	GalaxyOptions(),

	cScreen(CompositeScreen::get(s)),
	gScreen(GLScreen::get(s))
{
	ScreenInterface::setHandler(screen);
	CompositeScreenInterface::setHandler(cScreen);
	GLScreenInterface::setHandler(gScreen);

	/* create light, space and matrix */
	loadLightTexture();
	loadSpaceTexture();
	updateMatrix();

	/* create empty galaxy texture */
	galaxySize = CompSize(screen->width(), screen->height());
	galaxyTexture = GLTexture::imageBufferToTexture(NULL, galaxySize);

	if(!galaxyTexture.size())
	{
		/* This should never happen. If it happens, it will lead to a crash. */
		compLogMessage("galaxy", CompLogLevelFatal, "Cannot create galaxy texture!");
		return;
	}

	galaxyTexture[0]->setFilter(GL_LINEAR);
	galaxyTexture[0]->setWrap(GL_CLAMP_TO_EDGE);

	/* create galaxy buffer */
	if(GL::fbo)
	{
		GL::genFramebuffers(1, &galaxyBuffer);
		GL::bindFramebuffer(GL_FRAMEBUFFER_EXT, galaxyBuffer);
		GL::framebufferTexture2D(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT,
			GL_TEXTURE_2D, galaxyTexture[0]->name(), 0);
		GL::bindFramebuffer(GL_FRAMEBUFFER_EXT, 0);
	}

	/* galaxy toggle key */
	state = GalaxyScreen::GalaxyFalse;
	optionSetToggleKeyInitiate(boost::bind(&GalaxyScreen::initGalaxy,
				this, _1, _2, _3));
	optionSetToggleKeyTerminate(boost::bind(&GalaxyScreen::termGalaxy,
				this, _1, _2, _3));

	/* create particle system */
	ps = new SpiralGalaxy();
	ps->setStarsSize( optionGetStarsSize() );
	ps->generateFlare( optionGetDrawStreaks() );

	try
	{
		ps->setStarsCount( optionGetStarsCount() );
	}
	catch(std::bad_alloc&)
	{
		/*
		 * If this happens the plugin remains enabled but won't do anything.
		 * The user won't get any error message but cannot activate the galaxy.
		 */
		compLogMessage("galaxy", CompLogLevelFatal, "Cannot allocate enough memory for particle system!");

		state = GalaxyScreen::GalaxyUndefined;
	}

	/* particlesOptionChanged */
	optionSetStarsCountNotify(boost::bind(&GalaxyScreen::
				particlesOptionChanged, this, _1, _2));
	optionSetStarsSizeNotify(boost::bind(&GalaxyScreen::
				particlesOptionChanged, this, _1, _2));
	optionSetDrawStreaksNotify(boost::bind(&GalaxyScreen::
				particlesOptionChanged, this, _1, _2));

	/* matrixOptionChanged */
	optionSetOffsetXNotify(boost::bind(&GalaxyScreen::
				matrixOptionChanged, this, _1, _2));
	optionSetOffsetYNotify(boost::bind(&GalaxyScreen::
				matrixOptionChanged, this, _1, _2));

	optionSetZoomNotify(boost::bind(&GalaxyScreen::
				matrixOptionChanged, this, _1, _2));

	optionSetRotateXNotify(boost::bind(&GalaxyScreen::
				matrixOptionChanged, this, _1, _2));
	optionSetRotateYNotify(boost::bind(&GalaxyScreen::
				matrixOptionChanged, this, _1, _2));
	optionSetRotateZNotify(boost::bind(&GalaxyScreen::
				matrixOptionChanged, this, _1, _2));

	/* wallpaper atom */
	compizWallpaperAtom = XInternAtom(screen->dpy(),
		"_COMPIZ_WALLPAPER_SUPPORTED", 0);

	/*
	 * We set it permanent, because it doesn't depend on a setting if
	 * we can provide an background.
	*/
	unsigned char sd = 1;
	XChangeProperty(screen->dpy(), screen->root(), compizWallpaperAtom,
		XA_CARDINAL, 8, PropModeReplace, &sd, 1);

	/* create fake desktop window? */
	fakeDesktop = None;
	if(!screen->desktopWindowCount())
	{
		createFakeDesktopWindow();

		if(state != GalaxyScreen::GalaxyUndefined)
			state = GalaxyScreen::GalaxyTrue;
	}
}

GalaxyScreen::~GalaxyScreen()
{
	/* delete wallpaper atom */
	XDeleteProperty(screen->dpy(), screen->root(), compizWallpaperAtom);

	/* destroy fake desktop window */
	if(fakeDesktop != None)
		destroyFakeDesktopWindow();

	/* free galaxy buffer */
	if(GL::fbo)
		GL::deleteFramebuffers(1, &galaxyBuffer);

	/* free particle system */
	delete ps;
}

GalaxyWindow::GalaxyWindow(CompWindow *w):
	PluginClassHandler<GalaxyWindow, CompWindow>(w),

	window(w),
	cWindow(CompositeWindow::get(w)),
	gWindow(GLWindow::get(w))
{
	CompositeWindowInterface::setHandler(cWindow);
	GLWindowInterface::setHandler(gWindow);
}


bool
GalaxyPluginVTable::init()
{
	if( !CompPlugin::checkPluginABI("core", CORE_ABIVERSION) ||
		!CompPlugin::checkPluginABI("composite", COMPIZ_COMPOSITE_ABI) ||
		!CompPlugin::checkPluginABI("opengl", COMPIZ_OPENGL_ABI)
	)
		return false;

	return true;
}
