/**
 *
 * Compiz galaxy livewallpaper plugin
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright (C) 2011 Maximilian Schnarr <Fyrmir@gmx.de>
 *
 *
 * Several parts taken from the original plugin developed by Pal Dorogi
 * with the following copyright:
 *
 * Copyright (C) 2010 Pal Dorogi <pal.dorogi@gmail.com>
 *
 * Original plugin can be found here:
 * http://launchpad.net/compiz-plugins-livewallpapers
 *
 **/

#include <cmath>
#include <cstring>

#define GL_GLEXT_PROTOTYPES // glPointParameterfv
#include "particle.h"

/*
 * Set the color
 *
 * O     155 176 255  #9bb0ff  Blue
 * B     170 191 255  #aabfff  Blue-White
 * A     202 215 255  #cad7ff  White
 * F     248 247 255  #f8f7ff  White
 * G     255 244 234  #fff4ea  Yellow-White
 * K     255 210 161  #ffd2a1  Orange
 * M     255 204 111  #ffcc6f  Red
 * Reference: http://www.vendian.org/mncharity/dir3/starcolor/
 * Use my personal favorite mneomnic:
 * Only Bored Astronomers Find Gratification Knowing Mnemonics:)
 */
unsigned char star_colors[7][3] =
{
	{155, 176, 255},
	{170, 191, 255},
	{202, 215, 255},
	{248, 247, 255},
	{255, 244, 234},
	{255, 210, 161},
	{255, 204, 111}
};

enum _star_type {O, B, A, F, G, K, M} star_type;

#define RED 0
#define GREEN 1
#define BLUE 2

#define _PI  (float) 3.141593
#define _2PI (float) (2 * _PI)

/* 0 (line) -> 1.0f (circle) */
#define ELLIPSE_RATIO .885f

#define FLARE_SIZE 32

static inline float
randf()
{
	return  rand () / (float) RAND_MAX;
}

static inline float
rand1f(float num)
{
	return  num * randf ();
}

static inline float
rand2f(float min, float max)
{
    return min + (max - min) * randf ();
}

static inline float
clampf(float num, float min, float max)
{
   if(num > max)
   {
     num = max;
   }
   else if (num < min)
   {
     num = min;
   }

   return num;
}

/*
 * Gaussian random generator
 * Using the polar form of the Box-Muller transformation
 */
static float
randng()
{
    static float y1, y2;
    static int have_next = FALSE;

    if (have_next)
    {
        have_next = FALSE;
        return y2;
    }

    float x1, x2, w;
    do
    {
        x1 = 2.0 * randf () - 1.0;
        x2 = 2.0 * randf () - 1.0;
        w = x1 * x1 + x2 * x2;
    } while (w >= 1.0);

    w = sqrt(-2.0 * log (w) / w);
    y1 = x1 * w;
    y2 = x2 * w;
    have_next = TRUE;

    return y1;
}


Particle::Particle()
{
	angle = rand1f(_2PI);

	distance = fabs(randng()) * 0.5f;

	/*
	 * Phi depends on the distance of the star to get density waves.
	 * See http://en.wikipedia.org/wiki/Density_wave_theory
	 */
	cos_phi = cosf(distance * _2PI);
	sin_phi = sinf(distance * _2PI);

	z = (1 - distance) * 0.25 * randng() /
		(_PI * (1 + (2 * distance * distance)));

	if(distance < 0.2)
		z *= 0.8f;

	/*
	 * The angular velocity of the particle.
	 * Unit: 1/ms
	 */
	speed = - ( rand2f(0.0012f, 0.0018f) / distance * 0.5f ) / 20.0f;

	if(distance < 0.15f)
		star_type = B;
	else if(distance < 0.25f)
		star_type = A;
	else if(distance < 0.35f)
		star_type = F;
	else
		star_type = G;

	r = star_colors[star_type][RED];
	g = star_colors[star_type][GREEN];
	b = star_colors[star_type][BLUE];

	a = (GLubyte) clampf(255 - (1 - distance) * 100, 155, 255);
}

SpiralGalaxy::SpiralGalaxy():
	starsCount(0),
	starsSize(1),
	flareSize(FLARE_SIZE, FLARE_SIZE)
{}

void
SpiralGalaxy::setStarsCount(unsigned int count)
{
	stars.reserve(count);

	/* erase stars from the end */
	if(stars.size() > count)
		stars.erase(stars.begin() + count, stars.end());

	/* append stars */
	if(stars.size() < count)
		for(unsigned int i = stars.size(); i < count; i++)
			stars.push_back( Particle() );

	/* resize vertex and color vector */
	vertices.resize(3 * count + 1);
	colors.resize(4 * count + 1);

	starsCount = count;
}

unsigned int
SpiralGalaxy::getStarsCount()
{
	return starsCount;
}

void
SpiralGalaxy::setStarsSize(int size)
{
	starsSize = size;
}

int
SpiralGalaxy::getStarsSize()
{
	return starsSize;
}

void
SpiralGalaxy::generateFlare(bool drawStreaks)
{
	int w, h;
	int i, x, y;
	float dx, dy, r, R;
	float fa;

	w = FLARE_SIZE;
	h = FLARE_SIZE;
	R = (w + h) / 4;

	char *data = new char[w * h * 4];

	for(x = 0; x < w; x++)
	{
		for(y = 0; y < h; y++)
		{
			dx = (R - x) / R;
			dy = (R - y) / R;
			r = sqrt(dx * dx + dy * dy);

//			r = exp( -r );
//			r *= r;

			r = 1 - r;

			/* Draw flare */
			fa = r * r;
			if(r > 1) fa = 1;

			/* Draw streaks */
			if(drawStreaks && ((dx == 0.0f) || (dy == 0.0f)))
			{
				fa += r * 0.5;
				if(fa > 1) fa = 1;
			}

			if((x == 0) || (x == (FLARE_SIZE - 1)) ||
			   (y == 0) || (y == (FLARE_SIZE - 1)))
			{
				fa = 0;
			}

			/*
			 * Save color as GL_BGRA.
			 * See GLTexture::imageBufferToTexture
			 */
			i = ( x + y * w) * 4;
			data[i + 0] = 255; /* Blue */
			data[i + 1] = 255; /* Green */
			data[i + 2] = 255; /* Red */
			data[i + 3] = (char) (255 * fa); /* Alpha */
		}
	}

	flareTexture = GLTexture::imageBufferToTexture(data, flareSize);

	if(data) delete[] data;

	if(!flareTexture.size())
		return;

	flareTexture[0]->setFilter(GL_LINEAR);
	flareTexture[0]->setWrap(GL_CLAMP_TO_EDGE);
}

void
SpiralGalaxy::updateParticles(int msSinceLastPaint)
{
	static float a_cos_angle, b_sin_angle;
	static std::vector<Particle>::iterator p;
	static std::vector<GLfloat>::iterator v;
	static std::vector<GLubyte>::iterator c;

	v = vertices.begin();
	c = colors.begin();
	for(p = stars.begin(); p < stars.end(); p++)
	{
		/* Increase the angle. */
		p->angle += msSinceLastPaint * p->speed;

		/*
		 * a is the semi-major axis, here the distance and b is the
		 * semi-minor axis defined by the distance and the ellipse ratio.
		 */
		a_cos_angle = p->distance * cosf(p->angle);
		b_sin_angle = p->distance * ELLIPSE_RATIO * sinf(p->angle);

		/*
		 * To calculate the current position we use the general
		 * parametic form of an ellipse.
		 * See http://en.wikipedia.org/wiki/Ellipse#General_parametric_form
		 */
		v[0] = a_cos_angle * p->cos_phi - b_sin_angle * p->sin_phi;
		v[1] = a_cos_angle * p->sin_phi + b_sin_angle * p->cos_phi;
		v[2] = p->z;
		v += 3;

		/* Copy the color to the color vector. */
		c[0] = p->r;
		c[1] = p->g;
		c[2] = p->b;
		c[3] = p->a;
		c += 4;
	}
}

void
SpiralGalaxy::drawParticles(bool starByStar)
{
	static GLfloat quadratic[] = {0.4f, 0.0f, 0.1f};
	static GLfloat max_size = 0.0f;

	if(flareTexture.size())
		flareTexture[0]->enable(GLTexture::Good);

	glPointParameterfv(GL_POINT_DISTANCE_ATTENUATION, quadratic);
	glPointParameterf(GL_POINT_FADE_THRESHOLD_SIZE, 60.0f);

	glGetFloatv(GL_POINT_SIZE_MAX, &max_size);
	glPointParameterf(GL_POINT_SIZE_MIN, 0.1f);
	#ifdef POINT_SIZE_MAX_WORKS
	glPointParameterf(GL_POINT_SIZE_MAX, max_size);
	#endif

	glDepthMask(GL_FALSE);
	glPointSize(starsSize);

	glTexEnvf(GL_POINT_SPRITE_ARB, GL_COORD_REPLACE, GL_TRUE);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	glEnable(GL_POINT_SPRITE);

	if(!starByStar)
	{
		/* Render stars using vertex array. */
		glEnableClientState(GL_COLOR_ARRAY);

		glVertexPointer(3, GL_FLOAT, 3 * sizeof(GLfloat), &vertices[0]);
		glColorPointer(4, GL_UNSIGNED_BYTE, 4 * sizeof(GLubyte), &colors[0]);

		glDrawArrays(GL_POINTS, 0, starsCount);

		glDisableClientState(GL_COLOR_ARRAY);
	}
	else
	{
		/*
		 * Render stars vertex by vertex.
		 * This is only an alternative solution. On my system using
		 * Unity with fglrx driver compiz will crash on glDrawArrays.
		 * To solve this problem you shouldn't use the texture or
		 * glDrawArrays. I decided that replacing glDrawArrays is the
		 * best solution until this is fixed somehow.
		 */
		static std::vector<GLfloat>::iterator v;
		static std::vector<GLubyte>::iterator c;

		c = colors.begin();
		glBegin(GL_POINTS);

		for(v = vertices.begin(); v < vertices.end(); v += 3)
		{
			glColor4ub(c[0], c[1], c[2], c[3]);
			c += 4;

			glVertex3f(v[0], v[1], v[2]);
		}

		glEnd();
		glColor4usv(defaultColor);
	}

	glDisable(GL_POINT_SPRITE);

	if(flareTexture.size())
		flareTexture[0]->disable();
}
